using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Net.Http;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Extensions.Logging;
using Props.Shop.Framework;

namespace Props.Shop.Adafruit.Api
{
    public class LiveProductListingManager : IProductListingManager
    {
        private ILogger<LiveProductListingManager> logger;
        private bool disposedValue;
        private int minDownloadInterval;
        private DateTime? lastDownload;
        private object refreshLock = new object();
        private volatile Task<IDictionary<string, IList<ProductListing>>> productListingsTask;

        public Task<IDictionary<string, IList<ProductListing>>> ProductListings => productListingsTask;
        private readonly ConcurrentDictionary<string, ProductListing> activeProductListingUrls = new ConcurrentDictionary<string, ProductListing>();

        private ProductListingsParser parser = new ProductListingsParser();
        private HttpClient httpClient;
        private Timer updateTimer;

        public LiveProductListingManager(HttpClient httpClient, ILogger<LiveProductListingManager> logger, int minDownloadInterval = 5 * 60 * 1000)
        {
            this.logger = logger;
            this.minDownloadInterval = minDownloadInterval;
            this.httpClient = httpClient;
        }

        public void RefreshProductListings()
        {
            lock (refreshLock)
            {
                if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
                if ((lastDownload != null && DateTime.UtcNow - lastDownload <= TimeSpan.FromMilliseconds(minDownloadInterval)) || (productListingsTask != null && !productListingsTask.IsCompleted)) return;
                lastDownload = DateTime.UtcNow;
                logger.LogDebug("Refreshing listings ({0}).", lastDownload);
                productListingsTask = DownloadListings();
            }
        }

        public async Task<ProductListing> GetProductListingFromUrl(string url)
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            await productListingsTask;
            return activeProductListingUrls[url];
        }

        private async Task<IDictionary<string, IList<ProductListing>>> DownloadListings()
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            logger.LogDebug("Beginning listing database download.");
            HttpResponseMessage responseMessage = await httpClient.GetAsync("products");
            parser.BuildProductListings(responseMessage.Content.ReadAsStream());
            logger.LogDebug("Listing database parsed.");
            Dictionary<string, IList<ProductListing>> listingNames = new Dictionary<string, IList<ProductListing>>();
            activeProductListingUrls.Clear();
            foreach (ProductListing product in parser.ProductListings)
            {
                activeProductListingUrls.TryAdd(product.URL, product);
                IList<ProductListing> sameProducts = listingNames.GetValueOrDefault(product.Name);
                if (sameProducts == null)
                {
                    sameProducts = new List<ProductListing>();
                    listingNames.Add(product.Name, sameProducts);
                }

                sameProducts.Add(product);
            }
            logger.LogDebug("Downloaded listings organized.");
            return listingNames;
        }

        public void StartUpdateTimer(int delay = 1000 * 60 * 5, int period = 1000 * 60 * 5)
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (updateTimer != null) throw new InvalidOperationException("Update timer already started.");
            logger.LogInformation("Starting update timer.");
            updateTimer = new Timer((state) =>
            {
                RefreshProductListings();
            }, null, delay, period);
        }

        public void StopUpdateTimer()
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (updateTimer != null) throw new InvalidOperationException("Update timer not started.");
            logger.LogInformation("Stopping update timer.");
            updateTimer.Dispose();
            updateTimer = null;
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    updateTimer?.Dispose();
                    updateTimer = null;
                }
                disposedValue = true;
            }
        }

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
    }
}