using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using Props.Shop.Adafruit.Api;
using Props.Shop.Framework;

namespace Props.Shop.Adafruit.Tests.Api
{
    public class FakeProductListingManager : IProductListingManager
    {
        private Timer refreshTimer;
        private bool disposedValue;
        private volatile Task<IDictionary<string, IList<ProductListing>>> activeListings;
        private DateTime? lastDownload;
        private ProductListingsParser parser = new ProductListingsParser();
        private readonly ConcurrentDictionary<string, ProductListing> activeProductListingUrls = new ConcurrentDictionary<string, ProductListing>();


        public Task<IDictionary<string, IList<ProductListing>>> ProductListings => activeListings;

        public async Task<ProductListing> GetProductListingFromUrl(string url)
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            await activeListings;
            return activeProductListingUrls[url];
        }

        public void RefreshProductListings()
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if ((lastDownload != null && DateTime.UtcNow - lastDownload <= TimeSpan.FromMilliseconds(5 * 60 * 1000)) || (activeListings != null && !activeListings.IsCompleted)) return;
            activeListings = DownloadListings();
        }

        private Task<IDictionary<string, IList<ProductListing>>> DownloadListings() {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            lastDownload = DateTime.UtcNow;
            parser.BuildProductListings(File.OpenRead("./Assets/products.json"));
            Dictionary<string, IList<ProductListing>> listingNames = new Dictionary<string, IList<ProductListing>>();
            activeProductListingUrls.Clear();
            foreach (ProductListing product in parser.ProductListings)
            {
                activeProductListingUrls.TryAdd(product.URL, product);
                IList<ProductListing> sameProducts = listingNames.GetValueOrDefault(product.Name);
                if (sameProducts == null) {
                    sameProducts = new List<ProductListing>();
                    listingNames.Add(product.Name, sameProducts);
                }

                sameProducts.Add(product);
            }
            return Task.FromResult<IDictionary<string, IList<ProductListing>>>(listingNames);
        }

        public void StartUpdateTimer(int delay = 300000, int period = 300000)
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (refreshTimer != null) throw new InvalidOperationException("Refresh timer already running.");
            refreshTimer = new Timer((state) => {
                RefreshProductListings();
            }, null, delay, period);
        }

        public void StopUpdateTimer()
        {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (refreshTimer == null) throw new InvalidOperationException("Refresh timer not running.");
            refreshTimer.Dispose();
            refreshTimer = null;
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    refreshTimer?.Dispose();
                    refreshTimer = null;
                }

                disposedValue = true;
            }
        }

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
    }
}